#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <dirent.h>
#include <time.h>


#include <dinouml_com.h>

static short dinouml_debug = 0;

struct DinoUMLList {
	struct DinoUMLList	*next;
	char			*val;
};

struct DinoUMLIf {
	struct DinoUMLIf	*next;

	char			*name;
	short			type, subtype;
	short			mode;
	struct DinoUMLList	*ips;
	char			*cir;
	char			*mir;
	char			*qlen;
	char			*mtu;
	char			*tos;
	char			*delay;
	char			*drop;
	char			*dup;
	short			stp;
	char			*local, *remote;
	char			*ttl;
	struct DinoUMLList	*devs;
	char			*mac;
	char			*vlan_id;
	char			*vlan_dev;
	char			*vutp;
};

struct DinoUML {
	char			*base;
	char			*topo;

	struct DinoUML		*machines;

	char			*nice;
	char			*dir;
	char			*name;
	char			*kernel;
	char			*para;
	char			*root;
	char			*domain;
	struct DinoUMLList	*nameservers;
	long			mem;
	long			tmp;
	long			swap;
	char			*hosts;
	struct DinoUMLIf	*ifs;

	int			line;
	char			buf[4096];
	FILE			*f;
	short			buf_valid;
	char			*port;

	struct DinoUMLList	*run;
};



static void dinouml_copy(void)
{
	fprintf(stderr, "DinoUML - Copyright Catalin(ux aka Dino) BOIE - 2004-2006\n");
	fprintf(stderr, "\n");
}

static pid_t dinouml_fork(char *argv[], short waits, char *file1, char *file2, char *pid)
{
	pid_t	ret, w;
	int	status, fd1, fd2, fdx;

	if (dinouml_debug > 0) {
		fprintf(stderr, "DEBUG: Running ");
		fd1 = 0;
		while (argv[fd1]) {
			fprintf(stderr, "%s ", argv[fd1]);
			fd1++;
		}
		fprintf(stderr, "\n");
	}

	ret = fork();
	if (ret == -1) {
		fprintf(stderr, "Cannot fork [%s] [%s]!\n",
			argv[0], strerror(errno));
		exit(1);
	}

	if (ret == 0) { /* child */
		fdx = open("/dev/null", O_RDONLY);
		if (fdx == -1) {
			fprintf(stderr, "Cannot open /dev/null [%s]!\n",
				strerror(errno));
			exit(1);
		}
		dup2(fdx, 0);

		/*fprintf(stderr, "I'm the child and I will run cmd...\n");*/
		if (file1 != NULL) {
			fd1 = open(file1, O_WRONLY | O_CREAT | O_TRUNC, 0600);
			if (fd1 == -1) {
				fprintf(stderr, "Cannot create file [%s] for output [%s]!\n",
					file1, strerror(errno));
				exit(1);
			}
			dup2(fd1, 1);
		} else {
			dup2(fdx, 1);
		}

		if (file2 != NULL) {
			fd2 = open(file2, O_WRONLY | O_CREAT | O_TRUNC, 0600);
			if (fd2 == -1) {
				fprintf(stderr, "Cannot create file [%s] for output [%s]!\n",
					file2, strerror(errno));
				exit(1);
			}
			dup2(fd2, 2);
		} else {
			dup2(fdx, 2);
		}

		ret = execvp(argv[0], argv);
		if (ret == -1) {
			fprintf(stderr, "Cannot exec [%s] [%s]!\n",
				argv[0], strerror(errno));
			exit(1);
		}
	} else { /* parent */
		if (waits > 0) {
			/*fprintf(stderr, "Wainting for process %d to exit...\n", ret);*/
			w = wait(&status);
			/*
			fprintf(stderr, "The child %d exited with status %d [%s]!\n",
				w, status, strerror(errno));
			*/
		}

		if (pid != NULL) {
			int	fd;
			char	pid0[128];

			snprintf(pid0, sizeof(pid0), "%d", ret);

			fd = open(pid, O_WRONLY | O_CREAT | O_TRUNC, 0600);
			if (fd == -1) {
				fprintf(stderr, "Cannot write pid file [%s] [%s]!\n",
					pid, strerror(errno));
				exit(1);
			}

			write(fd, pid0, strlen(pid0));

			close(fd);
		}
	}

	if (dinouml_debug > 0)
		fprintf(stderr, "DEBUG:\tGot pid %d\n", ret);

	return ret;
}

static char *dinouml_get_random_mac(void)
{
	static char	ret[32];
	long int	r;

	r = random();

	snprintf(ret, sizeof(ret), "00:ee:%02lx:%02lx:%02lx:%02lx",
		(r >> 24) & 0xff,
		(r >> 16) & 0xff,
		(r >> 8) & 0xff,
		(r >> 0) & 0xff);

	return ret;
}

static int dinouml_build_file(char *file, long size)
{
	int		ret = 0;
	int		fd;
	struct stat	S;
	off_t		off;
	char		buf[4096];
	long		n, w;

	if (size == 0) {
		unlink(file);
	} else {
		off = 0;
		fd = open(file, O_WRONLY | O_CREAT | O_EXCL | O_APPEND, 0600);
		if (fd == -1) {
			if (errno == EEXIST) {
				ret = stat(file, &S);
				if (ret == -1) {
					fprintf(stderr, "Cannot stat file [%s]!\n",
						strerror(errno));
					return 13;
				}

				if (S.st_size == size) { /* size is ok, do nothing */
					return 0;
				}

				if (S.st_size > size)
					off = -1;

				fd = open(file, O_WRONLY | O_APPEND, 0600);
				if (fd == -1) {
					fprintf(stderr, "Cannot reopen file [%s] [%s]!\n",
						file, strerror(errno));
					return 19;
				}

				off = S.st_size;
			} else {
				fprintf(stderr, "Cannot create file [%s] [%s]!\n",
					file, strerror(errno));
				return 14;
			}
		}

		if (off == -1) { /* truncate */
			ret = ftruncate(fd, size);
			if (ret == -1) {
				fprintf(stderr, "Cannot truncate file [%s]!\n",
					strerror(errno));
				return 15;
			}
		} else { /* append data */
			memset(buf, 0, sizeof(buf));
			w = size - off;
			while (w > 0) {
				if (w > 4096)
					n = write(fd, buf, 4096);
				else
					n = write(fd, buf, w);
				if (n == -1) {
					fprintf(stderr, "Cannot append to file [%s]!\n",
						strerror(errno));
					return 16;
				}

				w -= n;
			}
		}

		close(fd);
	}

	return 0;
}

static int dinouml_build_dir(char *dir, char *last_dir)
{
	char		buf[8192];
	int		ret;
	struct stat	S;

	snprintf(buf, sizeof(buf),
		"%s/%s", dir, last_dir);

	ret = stat(buf, &S);
	if (ret == -1) {
		ret = mkdir(buf, 0655);
		if (ret == -1) {
			fprintf(stderr, "Cannot create machine dir [%s] [%s]!\n",
				buf, strerror(errno));
		}
	}

	return 0;
}

static int dinouml_build_swap(struct DinoUML *D)
{
	int	ret;
	char	file[8192];
	char	*argv[5];

	snprintf(file, sizeof(file), "%s/tmp/%s.SWAP",
		D->dir, D->name);
	ret = dinouml_build_file(file, D->swap * 1024 * 1024);
	if (ret < 0)
		return ret;

	argv[0] = "mkswap";
	argv[1] = file;
	argv[2] = NULL;
	dinouml_fork(argv, 1, NULL, NULL, NULL);

	return 0;
}

static char *dinouml_if_getmode(int mode)
{
	char	*ret;

	switch (mode) {
		case DINOUML_IF_MODE_SW: ret = "sw"; break;
		case DINOUML_IF_MODE_HUB: ret = "hub"; break;
		default:
			ret = "?";
	}

	return ret;
}

static char *dinouml_if_gettype(int type)
{
	char	*ret;

	switch (type) {
		case DINOUML_IF_TYPE_NORMAL: ret = "normal"; break;
		case DINOUML_IF_TYPE_BRIDGE: ret = "bridge"; break;
		case DINOUML_IF_TYPE_HOST: ret = "host"; break;
		case DINOUML_IF_TYPE_TUNNEL: ret = "tunnel"; break;
		case DINOUML_IF_TYPE_VLAN: ret = "vlan"; break;
		default:
			ret = "?";
	}

	return ret;
}

static void dinouml_host_ifs(struct DinoUML *D, FILE *out)
{
	struct DinoUMLIf	*If;

	If = D->ifs;
	while (If) {
		if (If->type == DINOUML_IF_TYPE_NORMAL) {
			fprintf(out, " %s=daemon,%s,,%s/tmp/%s",
				If->name, If->mac, D->dir, If->vutp);
		}
		If = If->next;
	}
}

static char *dinouml_if_build(struct DinoUMLIf *If)
{
	static char		tmp[256], buf[20480];
	int			size, ret;
	struct DinoUMLList	*dev, *ip;
	char			*parent, *root, limit[64];

	size = 0;
	strcpy(buf, "");

	switch (If->type) {
		case DINOUML_IF_TYPE_NORMAL: {
		} break;

		case DINOUML_IF_TYPE_HOST: {
		} break;

		case DINOUML_IF_TYPE_BRIDGE: {
			ret = snprintf(tmp, sizeof(tmp), "modprobe bridge\n");
			if (size + ret > sizeof(buf)) return NULL;
			strcat(buf, tmp);

			ret = snprintf(tmp, sizeof(tmp), "brctl addbr %s\n", If->name);
			if (size + ret > sizeof(buf)) return NULL;
			strcat(buf, tmp);

			dev = If->devs;
			while (dev) {
				ret = snprintf(tmp, sizeof(tmp), "brctl addif %s %s\n",
					If->name, dev->val);
				if (size + ret > sizeof(buf)) return NULL;
				strcat(buf, tmp);

				dev = dev->next;
			}

			if (If->stp == 1) {
				ret = snprintf(tmp, sizeof(tmp), "brctl stp %s on\n",
					If->name);
				if (size + ret > sizeof(buf)) return NULL;
				strcat(buf, tmp);
			}
		} break;

		case DINOUML_IF_TYPE_TUNNEL: {
			char	*type;

			switch (If->subtype) {
				case DINOUML_IF_SUBTYPE_GRE: type = "gre"; break;
				case DINOUML_IF_SUBTYPE_IPIP: type = "ipip"; break;
				case DINOUML_IF_SUBTYPE_SIT: type = "sit"; break;
				default:
					fprintf(stderr, "Invalid tunnel type [%d]!\n", If->subtype);
					return NULL;
			}

			ret = snprintf(tmp, sizeof(tmp), "modprobe ipip\nmodprobe ip_gre\nmodprobe ip6_tunnel\n");
			if (size + ret > sizeof(buf)) return NULL;
			strcat(buf, tmp);

			ret = snprintf(tmp, sizeof(tmp), "ip tunnel add %s mode %s%s%s%s%s%s%s%s%s\n",
				If->name, type,
				strlen(If->local) > 0 ? " local " : "",
				strlen(If->local) > 0 ? If->local : "",
				strlen(If->remote) > 0 ? " remote " : "",
				strlen(If->remote) > 0 ? If->remote : "",
				strlen(If->ttl) > 0 ? " ttl " : "",
				strlen(If->ttl) > 0 ? If->ttl : "",
				strlen(If->tos) > 0 ? " tos " : "",
				strlen(If->tos) > 0 ? If->tos : "");
			if (size + ret > sizeof(buf)) return NULL;
			strcat(buf, tmp);
		} break;

		case DINOUML_IF_TYPE_VLAN: {
			ret = snprintf(tmp, sizeof(tmp), "modprobe 8021q\n");
			if (size + ret > sizeof(buf)) return NULL;
			strcat(buf, tmp);

			ret = snprintf(tmp, sizeof(tmp), "vconfig add %s %s\n",
				If->vlan_dev, If->vlan_id);
			if (size + ret > sizeof(buf))  return NULL;
			strcat(buf, tmp);
		} break;

		default: {
			fprintf(stderr, "Invalid interface type [%d]!\n", If->type);
			return NULL;
		}
	}

	/* set link up */
	ret = snprintf(tmp, sizeof(tmp), "ip link set %s up\n", If->name);
	if (size + ret > sizeof(buf)) return NULL;
	strcat(buf, tmp);

	/* add ips */
	ip = If->ips;
	while (ip) {
		ret = snprintf(tmp, sizeof(tmp), "ip address add %s dev %s\n",
			ip->val, If->name);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);

		ip = ip->next;
	}

	/* set mtu */
	if (strlen(If->mtu) > 0) {
		ret = snprintf(tmp, sizeof(tmp), "ip link set %s mtu %s\n",
			If->name, If->mtu);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);
	}

	/* qos stuff */
	if (strlen(If->cir) > 0) {
		ret = snprintf(tmp, sizeof(tmp), "modprobe sch_htb\nmodprobe sch_netem\n");
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);

		ret = snprintf(tmp, sizeof(tmp), "tc qdisc add dev %s root handle 1: htb default 10\n",
			If->name);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);

		ret = snprintf(tmp, sizeof(tmp), "tc class add dev %s classid 1:10 parent 1: htb rate %s ceil %s\n",
			If->name, If->cir, If->mir);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);
		parent = " parent 1:10";
		root = "";
	} else {
		parent = "";
		root = " root";
	}

	if (strlen(If->qlen) > 0) {
		snprintf(limit, sizeof(limit), " limit %s", If->qlen);
	} else {
		strcpy(limit, "");
	}
	if (strlen(If->delay) > 0) {
		ret = snprintf(tmp, sizeof(tmp), "tc qdisc add dev %s%s%s netem%s delay %s\n",
			If->name, root, parent, limit, If->delay);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);
	} else if (strlen(If->drop) > 0) {
		ret = snprintf(tmp, sizeof(tmp), "tc qdisc add dev %s%s%s netem%s drop %s\n",
			If->name, root, parent, limit, If->drop);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);
	} else if (strlen(If->dup) > 0) {
		ret = snprintf(tmp, sizeof(tmp), "tc qdisc add dev %s%s%s netem%s duplicate %s\n",
			If->name, root, parent, limit, If->dup);
		if (size + ret > sizeof(buf)) return NULL;
		strcat(buf, tmp);
	}

	return buf;
}

static void dinouml_string_fix(char *buf)
{
	char	*s;
	int	len;

	s = buf;
	/* cut comments */
	while (*s != '\0') {
		if (*s == '#') {
			*s = '\0';
			break;
		}
		s++;
	}
	len = strlen(buf);
	if (len == 0)
		return;

	/* cut \n & \r */
	while ((buf[len - 1] != '\0') && strchr("\r\n \t", buf[len - 1])) {
		buf[len - 1] = '\0';
		len--;
	}
}

static int dinouml_if_add_ip(struct DinoUMLIf *If, char *ip)
{
	struct DinoUMLList	*p, *q;

	q = malloc(sizeof(struct DinoUMLList));
	if (q == NULL)
		return -1;
	memset(q, 0, sizeof(struct DinoUMLList));

	if (If->ips == NULL) {
		If->ips = q;
	} else {
		p = If->ips;
		while (p->next)
			p = p->next;
		p->next = q;
	}

	q->val = strdup(ip);
	q->next = NULL;

	return 0;
}

static int dinouml_if_add_dev(struct DinoUMLIf *If, char *dev)
{
	struct DinoUMLList	*p, *q;

	q = malloc(sizeof(struct DinoUMLList));
	if (q == NULL)
		return -1;
	memset(q, 0, sizeof(struct DinoUMLList));

	if (If->devs == NULL) {
		If->devs = q;
	} else {
		p = If->devs;
		while (p->next)
			p = p->next;
		p->next = q;
	}

	q->val = strdup(dev);
	q->next = NULL;

	return 0;
}

static int dinouml_nameserver_add(struct DinoUML *D, char *s)
{
	struct DinoUMLList	*p, *q;

	q = malloc(sizeof(struct DinoUMLList));
	if (q == NULL)
		return -1;
	memset(q, 0, sizeof(struct DinoUMLList));

	if (D->nameservers == NULL)
		D->nameservers = q;
	else {
		p = D->nameservers;
		while (p->next)
			p = p->next;
		p->next = q;
	}

	q->val = strdup(s);

	return 0;
}

static int dinouml_parse_run(struct DinoUML *D)
{
	struct DinoUMLList	*l, *p;
	char			*r, *s;

	D->buf_valid = 0;

	while (1) {
		if (D->buf_valid == 0) {
			r = fgets(D->buf, sizeof(D->buf), D->f);
			if (r == NULL)
				break;

			D->line++;
			D->buf_valid = 1;
		}

		dinouml_string_fix(D->buf);

		if (strlen(D->buf) == 0) {
			D->buf_valid = 0;
			continue;
		}

		/* end of section detected */
		if ((D->buf[0] != ' ') && (D->buf[0] != '\t')) {
			/*fprintf(stderr, "End of section detected buf=[%s]\n", D->buf);*/
			D->buf_valid = 1;
			break;
		}

		/* skip spaceses */
		s = D->buf;
		while ((*s != '\0') && ((*s == ' ') || (*s == '\t'))) s++;

		/* add to buffer */
		l = malloc(sizeof(struct DinoUMLList));
		if (l == NULL)
			return -1;
		memset(l, 0, sizeof(struct DinoUMLList));

		l->val = strdup(s);
		l->next = NULL;

		if (D->run == NULL)
			D->run = l;
		else {
			p = D->run;
			while (p->next)
				p = p->next;
			p->next = l;
		}

		D->buf_valid = 0;
	}

	return 0;
}

static int dinouml_parse_interface(struct DinoUML *D)
{
	char			*r, *s, var[128];
	int			ret = -1, i, _r;
	struct DinoUMLIf	*If, *p;

	If = malloc(sizeof(struct DinoUMLIf));
	if (If == NULL) {
		fprintf(stderr, "Cannot alloc memory!\n");
		ret = -1;
		goto out;
	}
	memset(If, 0, sizeof(struct DinoUMLIf));

	/* ignore 'interface' */
	s = D->buf;
	while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
	if (*s != '\0') s++;

	i = 0;
	while ((i < sizeof(var)) && (D->buf[i] != '\0') && (D->buf[i] != ' ') && (D->buf[i] != '\t'))
		var[i++] = *s++;
	var[i] = '\0';

	if (strlen(var) == 0) {
		fprintf(stderr, "Interface must have a name in line %d!\n",
			D->line);
		ret = -1;
		goto out;
	}

	D->buf_valid = 0;
	If->name = strdup(var);
	If->type = DINOUML_IF_TYPE_NORMAL;
	If->mode = DINOUML_IF_MODE_SW;
	If->cir = "";
	If->mir = "";
	If->qlen = "";
	If->mtu = "";
	If->tos = "";
	If->delay = "";
	If->drop = "";
	If->dup = "";
	If->stp = 1;
	If->local = "";
	If->remote = "";
	If->ttl = "64";
	If->mac = "";
	If->vutp = "";

	while(1) {
		if (D->buf_valid == 0) {
			r = fgets(D->buf, sizeof(D->buf), D->f);
			if (r == NULL)
				break;

			D->line++;
			D->buf_valid = 1;
		}

		dinouml_string_fix(D->buf);

		if (strlen(D->buf) == 0) {
			D->buf_valid = 0;
			continue;
		}

		/* end of section detected */
		if ((D->buf[0] != ' ') && (D->buf[0] != '\t')) {
			/*fprintf(stderr, "End of section detected buf=[%s]\n", D->buf);*/
			D->buf_valid = 1;
			break;
		}

		/* skip spaceses */
		s = D->buf;
		while ((*s != '\0') && ((*s == ' ') || (*s == '\t'))) s++;

		/*fprintf(stderr, "Read2: [%s]\n", s);*/

		/* get var */
		i = 0;
		while ((*s != '\0') && (*s != ' ') && (*s != '\t') && (i < sizeof(var)))
			var[i++] = *s++;
		var[i] = '\0';

		/* remove spaces */
		while ((*s != '\0') && ((*s == ' ') || (*s == '\t'))) s++;

		if (strlen(s) == 0) {
			fprintf(stderr, "Empty line detected\n");
			break;
		}

		switch (var[0]) {
			case 'a': /* add */
				_r = dinouml_if_add_dev(If, s);
				if (_r == -1) {
					fprintf(stderr, "Cannot add dev to interface!\n");
					ret = -1;
					goto out;
				}
				break;
			case 'c':
				switch (var[1]) {
					case 'i': /* cir */
						If->cir = strdup(s);
						break;
					default:
						fprintf(stderr, "Invalid var name [%s] in line %d!\n",
							var, D->line);
						ret = -1;
						goto out;
				}
				break;
			case 'd':
				switch (var[1]) {
					case 'e': /* delay */
						If->delay = strdup(s);
						break;
					case 'r': /* drop */
						If->drop = strdup(s);
						break;
					case 'u': /* dup */
						If->dup = strdup(s);
						break;
					default:
						fprintf(stderr, "Invalid var name [%s] in line %d!\n",
							var, D->line);
						ret = -1;
						goto out;
				}
				break;
			case 'i':
				_r = dinouml_if_add_ip(If, s);
				if (_r == -1) {
					fprintf(stderr, "Cannot add ip to interface in line %d!\n",
						D->line);
					ret = -1;
					goto out;
				}
				break;
			case 'l':
				If->local = strdup(s);
				break;
			case 'm':
				switch (var[1]) {
					case 'a': /* mac */
						If->mac = strdup(s);
						break;
					case 'i': /* mir */
						If->mir = strdup(s);
						break;
					case 'o': /* mode */
						switch (*s) {
							case 'h': If->mode = DINOUML_IF_MODE_HUB; break;
							case 's': If->mode = DINOUML_IF_MODE_SW; break;
							default:
								fprintf(stderr, "Invalid mode in line %d!\n",
									D->line);
								ret = -1;
								goto out;
						}
						break;
					case 't': /* mtu */
						If->mtu = strdup(s);
						break;
					default:
						fprintf(stderr, "Invalid var name [%s] in line %d!\n",
							var, D->line);
						ret = -1;
						goto out;
				}
				break;
			case 'q': /* qlen */
				If->qlen = strdup(s);
				break;
			case 'r': /* remote */
				If->remote = strdup(s);
				break;
			case 's':
				switch (var[1]) {
					case 't': /* stp */
						if ((*s == 'o') && (*(s+1) == 'f'))
							If->stp = 0;
						break;
					case 'u': /* subtype */
						switch (*s) {
							case 'g': If->subtype = DINOUML_IF_SUBTYPE_GRE; break;
							case 'i': If->subtype = DINOUML_IF_SUBTYPE_IPIP; break;
							case 's': If->subtype = DINOUML_IF_SUBTYPE_SIT; break;
							default:
								fprintf(stderr, "Invalid subtype in line %d!\n",
									D->line);
								ret = -1;
								goto out;
						}
						break;
					default:
						fprintf(stderr, "Invalid var name [%s] in line %d!\n",
							var, D->line);
						ret = -1;
						goto out;
				}
				break;
			case 't':
				switch (var[1]) {
					case 'y': /* type */
						switch (*s) {
							case 'n': If->type = DINOUML_IF_TYPE_NORMAL; break;
							case 'h': If->type = DINOUML_IF_TYPE_HOST; break;
							case 't': If->type = DINOUML_IF_TYPE_TUNNEL; break;
							case 'b': If->type = DINOUML_IF_TYPE_BRIDGE; break;
							case 'v': If->type = DINOUML_IF_TYPE_VLAN; break;
							default:
								fprintf(stderr, "Invalid interface type [%s] in line %d!\n",
									s, D->line);
								ret = -1;
								goto out;
						}
						break;
					case 'o': /* tos */
						If->tos = strdup(s);
						break;
					case 't': /* ttl */
						If->ttl = strdup(s);
						break;
					default:
						fprintf(stderr, "Invalid var name [%s] in line %d!\n",
							var, D->line);
						ret = -1;
						goto out;
				}
				break;
			case 'v':
				If->vutp = strdup(s);
				break;
			default:
				fprintf(stderr, "Invalid var name [%s] in line %d!\n",
					var, D->line);
				ret = -1;
				goto out;
		}

		D->buf_valid = 0;
	}
	If->next = NULL;

	/* fix something */
	if (strlen(If->mir) == 0)
		If->mir = strdup(If->cir);
	if (((If->type == DINOUML_IF_TYPE_NORMAL) || (If->type == DINOUML_IF_TYPE_HOST)) && (strlen(If->mac) == 0))
		If->mac = strdup(dinouml_get_random_mac());
	/* fix vlan */
	if (If->type == DINOUML_IF_TYPE_VLAN) {
		char *_s;

		_s = strrchr(If->name, '.');
		if (_s) {
			*_s = '\0';
			_s++;
			If->vlan_id = strdup(_s);
			If->vlan_dev = strdup(If->name);
			_s--;
			*_s = '.';
		}
	}

	/* add if to structure */
	if (D->ifs == NULL) {
		D->ifs = If;
	} else {
		p = D->ifs;
		while (p) {
			if (strcmp(p->name, If->name) == 0) {
				fprintf(stderr, "Duplicate interface [%s] in line %d\n",
					If->name, D->line);
				ret = -1;
				goto out;
			}
			p = p->next;
		}

		p = D->ifs;
		while(p->next)
			p = p->next;
		p->next = If;
	}

	ret = 0;

	out:
	return ret;
}

static struct DinoUML *dinouml_parse(struct DinoUML *parent, char *topology, char *machine)
{
	int			ret;
	char			*r, *s, tmp[1024], file[4096];
	struct DinoUML		*D;
	struct DinoUMLList	*l;

	snprintf(file, sizeof(file), "%s/%s.conf",
		topology, machine);

	D = malloc(sizeof(struct DinoUML));
	if (!D) {
		fprintf(stderr, "Cannot alloc memory!\n");
		return NULL;
	}
	memset(D, 0, sizeof(struct DinoUML));

	D->dir = strdup(topology);
	D->topo = strdup(topology);

	D->f = fopen(file, "r");
	if (D->f == NULL) {
		fprintf(stderr, "Cannot open %s [%s]!\n",
			file, strerror(errno));
		return NULL;
	}

	/* set defaults from parent if available. If not use some default. */
	D->name = strdup(machine);
	D->line = 0;
	D->buf_valid = 0;
	D->hosts = "";
	D->base = strdup("/data/DinoUML");
	if (parent) {
		D->kernel = strdup(parent->kernel);
		D->root = strdup(parent->root);
		D->mem = parent->mem;
		D->tmp = parent->tmp;
		D->swap = parent->swap;
		D->para = parent->para;
		D->domain = parent->domain;
		D->nice = parent->nice;
		l = parent->nameservers;
		while (l) {
			dinouml_nameserver_add(D, l->val);
			l = l->next;
		}
		D->run = parent->run;
	} else {
		D->kernel = "kernels/default";
		D->root = "roots/default";
		D->mem = 64;
		D->tmp = 0;
		D->swap = 64;
		D->para = "";
		D->domain = "";
		D->nice = "0";
	}

	while (1) {
		if (D->buf_valid == 0) {
			r = fgets(D->buf, sizeof(D->buf), D->f);
			if (r == NULL)
				break;

			D->line++;
			D->buf_valid = 1;
		}

		dinouml_string_fix(D->buf);

		if (strlen(D->buf) == 0) {
			D->buf_valid = 0;
			continue;
		}

		/*fprintf(stderr, "Read: [%s]\n", D->buf);*/

		s = D->buf;
		if (strncmp(s, "base", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			free(D->base);
			D->base = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "run", 3) == 0) {
			ret = dinouml_parse_run(D);
			if (ret == -1)
				return NULL;
		} else if (strncmp(s, "kernel", 6) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->kernel = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "nice", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->nice = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "port", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}
			D->port = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "domain", 6) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}
			D->domain = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "nameserver", 10) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}
			dinouml_nameserver_add(D, s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "para", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			if (parent == NULL)
				D->para = strdup(s + 1);
			else {
				snprintf(tmp, sizeof(tmp),
					"%s %s", D->para, s + 1);
				D->para = strdup(tmp);
			}
			D->buf_valid = 0;
		} else if (strncmp(s, "name", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			free(D->name);
			D->name = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "root", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->root = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "mem", 3) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->mem = strtol(s + 1, NULL, 0);
			D->buf_valid = 0;
		} else if (strncmp(s, "tmp", 3) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->tmp = strtol(s + 1, NULL, 0);
			D->buf_valid = 0;
		} else if (strncmp(s, "swap", 4) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->swap = strtol(s + 1, NULL, 0);
			D->buf_valid = 0;
		} else if (strncmp(s, "hosts", 5) == 0) {
			while ((*s != '\0') && (*s != ' ') && (*s != '\t')) s++;
			if (*s == '\0') {
				fprintf(stderr, "Empty option in line %d.\n", D->line);
				return NULL;
			}

			D->hosts = strdup(s + 1);
			D->buf_valid = 0;
		} else if (strncmp(s, "interface", 9) == 0) {
			ret = dinouml_parse_interface(D);
			if (ret == -1)
				return NULL;
		} else {
			fprintf(stderr, "Parse error in line %d [%s] char=[%c]!\n",
				D->line, D->buf, *s);
			return NULL;
		}
	}

	fclose(D->f);

	/* add it to parent */
	if (parent) {
		while (parent->machines)
			parent = parent->machines;
		parent->machines = D;
	}

	return D;
}

static int dinouml_if_dump(struct DinoUMLIf *p)
{
	struct DinoUMLList	*ip, *dev;

	fprintf(stderr, "\tinterface %s\n", p->name);
	if (p->type > 0)
		fprintf(stderr, "\t\ttype %s\n", dinouml_if_gettype(p->type));
	if (p->subtype > 0)
		fprintf(stderr, "\t\tsubtype %d\n", p->subtype);
	if (p->mode > 0)
		fprintf(stderr, "\t\tmode %s\n", dinouml_if_getmode(p->mode));

	if (strlen(p->vutp) > 0)
		fprintf(stderr, "\t\tvutp %s\n", p->vutp);

	if (p->type == DINOUML_IF_TYPE_NORMAL) {
		if (strlen(p->mac) > 0)
			fprintf(stderr, "\t\tMAC %s\n", p->mac);
	}

	if (p->type == DINOUML_IF_TYPE_BRIDGE) {
		if (p->stp == 1)
			fprintf(stderr, "\t\tstp on\n");
	}

	if (p->type == DINOUML_IF_TYPE_TUNNEL) {
		if (strlen(p->tos) > 0)
			fprintf(stderr, "\t\ttos %s\n", p->tos);
		if (strlen(p->local) > 0)
			fprintf(stderr, "\t\tlocal %s\n", p->local);
		if (strlen(p->remote) > 0)
			fprintf(stderr, "\t\tremote %s\n", p->remote);
		if (strlen(p->ttl) > 0)
			fprintf(stderr, "\t\tttl %s\n", p->ttl);
	}

	if (p->type == DINOUML_IF_TYPE_VLAN) {
		if (strlen(p->vlan_dev) > 0)
			fprintf(stderr, "\t\tvlan_dev %s\n", p->vlan_dev);
		if (strlen(p->vlan_id) > 0)
			fprintf(stderr, "\t\tvlan_id %s\n", p->vlan_id);
	}

	if (strlen(p->mtu) > 0)
		fprintf(stderr, "\t\tmtu %s\n", p->mtu);

	if (strlen(p->cir) > 0) {
		fprintf(stderr, "\t\tcir %s\n", p->cir);
		fprintf(stderr, "\t\tmir %s\n", p->mir);
	}
	if (strlen(p->qlen) > 0)
		fprintf(stderr, "\t\tqlen %s\n", p->qlen);
	if (strlen(p->drop) > 0)
		fprintf(stderr, "\t\tdrop %s\n", p->drop);
	if (strlen(p->dup) > 0)
		fprintf(stderr, "\t\tdup %s\n", p->dup);
	if (strlen(p->delay) > 0)
		fprintf(stderr, "\t\tdelay %s\n", p->delay);

	/* ips */
	ip = p->ips;
	while (ip) {
		fprintf(stderr, "\t\tip %s\n", ip->val);
		ip = ip->next;
	}

	/* devs */
	if (p->type == DINOUML_IF_TYPE_BRIDGE) {
		dev = p->devs;
		while (dev) {
			fprintf(stderr, "\t\tadd %s\n", dev->val);
			dev = dev->next;
		}
	}

	fprintf(stderr, "\n");

	return 0;
}

static int dinouml_dump(struct DinoUML *D)
{
	struct DinoUMLIf	*p;
	struct DinoUML		*m;
	struct DinoUMLList	*l;

	/* dump general data */
	fprintf(stderr, "\n");
	fprintf(stderr, "o Dump topology...\n");
	fprintf(stderr, "base %s\n", D->base);
	fprintf(stderr, "topology %s\n", D->topo);
	fprintf(stderr, "dir %s\n", D->dir);
	fprintf(stderr, "default kernel = [%s]\n", D->kernel);
	fprintf(stderr, "default para = [%s]\n", D->para);
	fprintf(stderr, "default root = [%s]\n", D->root);
	fprintf(stderr, "default mem = [%ld]\n", D->mem);
	fprintf(stderr, "default tmp = [%ld]\n", D->tmp);
	fprintf(stderr, "default swap = [%ld]\n", D->swap);
	fprintf(stderr, "default domain = [%s]\n", D->domain);
	fprintf(stderr, "default nameservers = ");
	l = D->nameservers;
	while (l) {
		fprintf(stderr, "%s ", l->val);
		l = l->next;
	}
	fprintf(stderr, "\n");
	fprintf(stderr, "hosts = [%s]\n", D->hosts);

	if (D->ifs->type == DINOUML_IF_TYPE_HOST)
		dinouml_if_dump(D->ifs);

	fprintf(stderr, "\n");
	fprintf(stderr, "o Dump machines...\n");
	m = D->machines;
	while (m) {
		fprintf(stderr, "[%s]\n", m->name);
		fprintf(stderr, "kernel = [%s]\n", m->kernel);
		fprintf(stderr, "para = [%s]\n", m->para);
		fprintf(stderr, "root = [%s]\n", m->root);
		fprintf(stderr, "mem = [%ld]\n", m->mem);
		fprintf(stderr, "tmp = [%ld]\n", m->tmp);
		fprintf(stderr, "swap = [%ld]\n", m->swap);
		fprintf(stderr, "nice = [%s]\n", m->nice);
		fprintf(stderr, "domain = [%s]\n", D->domain);
		fprintf(stderr, "nameservers = ");
		fprintf(stderr, "run = \n");
		l = m->run;
		while (l) {
			fprintf(stderr, "\t%s\n", l->val);
			l = l->next;
		}

		l = m->nameservers;
		while (l) {
			fprintf(stderr, "%s ", l->val);
			l = l->next;
		}
		fprintf(stderr, "\n");
		fprintf(stderr, "port = [%s]\n", m->port);

		p = m->ifs;
		while (p) {
			dinouml_if_dump(p);
			p = p->next;
		}

		m = m->machines;
	}

	return 0;
}

/* test if a uml_switch is up */
static int dinouml_switch_check(char *path)
{
	int			fd, ret, R = 0;
	struct sockaddr_un	sun;

	if (dinouml_debug > 0)
		fprintf(stderr, "DEBUG: %s: check if %s is alive...\n",
			__FUNCTION__, path);

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd == -1) {
		fprintf(stderr, "ERROR: Cannot create a socket [%s]!\n",
			strerror(errno));
		exit(1);
	}

	sun.sun_family = AF_UNIX;
	snprintf(sun.sun_path, sizeof(sun.sun_path), "%s", path);

	ret = connect(fd, (struct sockaddr *) &sun, sizeof(sun));
	if (ret == -1) {
		if (dinouml_debug > 0)
			fprintf(stderr, "DEBUG:\tCannot connect to [%s] [%s]\n",
				path, strerror(errno));
		R = -1;
		goto out;
	}

	close(fd);

	out:
	if (dinouml_debug > 0)
		fprintf(stderr, "DEBUG:\t%s is %salive\n", path, (R == -1) ? "not " : "");
	return R;
}

/* delete all sockets from tmp dir */
static void dinouml_switch_stopall(struct DinoUML *D)
{
	char			path[8192], pid[128], pidpath[8192];
	DIR			*dir;
	struct dirent		*de;
	struct stat		S;
	int			fd, ret;
	off_t			n;
	pid_t			ipid;
	int			alive, tries;

	fprintf(stderr, "o Stopping all switches...\n");

	snprintf(path, sizeof(path), "%s/tmp", D->dir);
	dir = opendir(path);
	if (!dir) {
		fprintf(stderr, "\nCannot open tmp dir [%s] [%s]!",
			path, strerror(errno));
		exit(1);
	}

	while (1) {
		de = readdir(dir);
		if (de == NULL)
			break;

		snprintf(path, sizeof(path), "%s/tmp/%s", D->dir, de->d_name);
		ret = stat(path, &S);
		if (ret == -1)
			continue;

		/* is a socket */
		if (!S_ISSOCK(S.st_mode))
			continue;

		alive = dinouml_switch_check(path);
		if (alive == -1)
			continue;

		/* stop process */
		snprintf(pidpath, sizeof(pidpath), "%s/tmp/%s.pid", D->dir, de->d_name);
		fd = open(pidpath, O_RDONLY);
		if (fd == -1) {
			fprintf(stderr, "\nCannot open pid [%s] [%s]!\n",
				pidpath, strerror(errno));
			continue;
		}

		n = read(fd, pid, sizeof(pid));
		if (n <= 0) {
			fprintf(stderr, "\nCannot read pid from file [%s] [%s]!",
				path, strerror(errno));
			exit(1);
		}
		pid[n] = '\0';

		close(fd);

		if (dinouml_debug > 0)
			fprintf(stderr, "DEBUG: Kill %s...\n", pid);
		ipid = strtol(pid, NULL, 0);
		kill(ipid, SIGKILL);

		tries = 100;
		while (tries > 0) {
			alive = dinouml_switch_check(path);
			if (alive == -1)
				break;

			usleep(100000);

			tries--;
		}

		unlink(path);

		if (tries == 0)
			fprintf(stderr, "WARNING: Give up waiting uml_switch to stop!\n");
	}

	closedir(dir);

	fprintf(stderr, "Done stopping all switches!\n");
	sleep(1);
}

static void dinouml_switch_start(struct DinoUMLIf *p, char *dir)
{
	char	path[8192], pid[8192], log1[8192], log2[8192];
	char	lock[8192];
	short	i;
	char	*argv[10];
	int	ret, tries;
	int	fd;
	char	*conf;
	int	lockfd, lockret;

	if ((p->type != DINOUML_IF_TYPE_NORMAL) && (p->type != DINOUML_IF_TYPE_HOST))
		return;

	if (strlen(p->vutp) == 0)
		return;

	fprintf(stderr, "o Try to start switch for vutp[%s] if[%s]: ",
		p->vutp, p->name);

	/* lock */
	snprintf(lock, sizeof(lock), "%s/tmp/%s.lock", dir, p->vutp);
	lockfd = open(lock, O_WRONLY | O_CREAT | O_TRUNC, 0600);
	if (lockfd == -1) {
		fprintf(stderr, "Cannot open file [%s] [%s]!\n",
			lock, strerror(errno));
		exit(1);
	}

	lockret = flock(lockfd, LOCK_EX);
	if (lockret == -1) {
		fprintf(stderr, "Cannot lock file [%s] [%s]!\n",
			lock, strerror(errno));
		exit(1);
	}

	snprintf(path, sizeof(path), "%s/tmp/%s", dir, p->vutp);

	/* already running? */
	ret = dinouml_switch_check(path);
	if (ret == 0) {
		lockret = flock(lockfd, LOCK_UN);
		if (lockret == -1) {
			fprintf(stderr, "Cannot unlock file [%s] [%s]!\n",
				lock, strerror(errno));
			exit(1);
		}
		close(lockfd);
		fprintf(stderr, " Already running.\n");
		return;
	}

	i = 0;
	argv[i++] = "uml_switch";
	argv[i++] = "-unix";
	argv[i++] = path;
	if (p->type == DINOUML_IF_TYPE_HOST) {
		argv[i++] = "-tap";
		argv[i++] = p->name;
	}
	if (p->mode == DINOUML_IF_MODE_HUB)
		argv[i++] = "-hub";
	argv[i++] = NULL;

	snprintf(log1, sizeof(log1), "%s/tmp/%s.log1", dir, p->vutp);
	snprintf(log2, sizeof(log1), "%s/tmp/%s.log2", dir, p->vutp);
	snprintf(pid, sizeof(pid), "%s/tmp/%s.pid", dir, p->vutp);

	dinouml_fork(argv, 0, log1, log2, pid);

	/* wait to start */
	tries = 100;
	while (tries > 0) {
		ret = dinouml_switch_check(path);
		if (ret == 0)
			break;

		usleep(100000);
		tries--;
	}
	if (tries == 0)
		fprintf(stderr, "WARNING: Give up waiting for uml_switch to start!\n");

	lockret = flock(lockfd, LOCK_UN);
	if (lockret == -1) {
		fprintf(stderr, "Cannot unlock file [%s] [%s]!\n",
			lock, strerror(errno));
		exit(1);
	}
	close(lockfd);

	fprintf(stderr, " Done.\n");

	/* configure host dev */
	if (p->type == DINOUML_IF_TYPE_HOST) {
		conf = dinouml_if_build(p);
		if (conf) {
			snprintf(path, sizeof(path), "%s/tmp/main.sh", dir);
			fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0700);
			if (fd > -1) {
				write(fd, "#!/bin/sh\n\n", 11);
				write(fd, conf, strlen(conf));
				close(fd);

				i = 0;
				argv[i++] = "sh";
				argv[i++] = "-c";
				argv[i++] = path;
				argv[i++] = NULL;
				dinouml_fork(argv, 1, NULL, NULL, NULL);
			}
		}
	}
}

static int dinouml_run(struct DinoUML *D)
{
	struct DinoUML		*m;
	struct DinoUMLIf	*p;
	struct DinoUMLList	*l;
	char			*s;
	FILE			*f, *ssf;
	char			buf[8192], tmp[8192], ss[8192];
	char			log1[8192], log2[8192];
	char			*argv[200];
	int			i, n;
	char			umlpid[8192];
	struct stat S;
	time_t root_mtime, cow_mtime;
	char root_tmp[4096];

	/* create tree */
	dinouml_build_dir(D->dir, "tmp");
	dinouml_build_dir(D->dir, "overlay");

	dinouml_switch_stopall(D);

	fprintf(stderr, "o Start switches...\n");
	if (D->ifs) {
		p = D->ifs;
		dinouml_switch_start(p, D->dir);
	}

	fprintf(stderr, "o Pack common overlay fs...\n");
	snprintf(buf, sizeof(buf),
		"tar -C%s/overlay -cf \"%s/tmp/overlay.tar\" .",
		D->dir, D->dir);
	system(buf); /* ??? replace with fork */

	fprintf(stderr, "o Start machines...\n");
	m = D->machines;
	while (m) {
		/* tree build */
		snprintf(buf, sizeof(buf), "overlay_%s", m->name);
		dinouml_build_dir(D->dir, buf);

		snprintf(buf, sizeof(buf), "%s/overlay_%s/DinoUML.sh", D->dir, m->name);
		f = fopen(buf, "w");
		if (f == NULL) {
			fprintf(stderr, "Cannot create %s file [%s]!\n",
				buf, strerror(errno));
			return -1;
		}

		fprintf(f, "#!/bin/sh\n\n");
		fprintf(f, "# This is an auto generated file. Do not edit!\n\n");

		/* i'm alive */
		fprintf(f, "echo \"%s-%s it's alive!\"\n\n",
			D->topo, m->name);

		/* set hostname */
		fprintf(f, "echo \"%s-%s\" > /proc/sys/kernel/hostname\n",
			D->topo, m->name);
		fprintf(f, "echo \"127.0.0.1	localhost %s-%s\" > /etc/hosts\n",
			D->topo, m->name);
		fprintf(f, "echo \"%s-%s\" > /etc/HOSTNAME\n",
			D->topo, m->name);
		fprintf(f, "\n");

		/* resolv.conf */
		if (strlen(m->domain) > 0)
			fprintf(f, "echo \"search %s\" > /etc/resolv.conf\n",
				m->domain);
		l = m->nameservers;
		while(l) {
			fprintf(f, "echo \"nameserver %s\" >> /etc/resolv.conf\n", l->val);
			l = l->next;
		}
		fprintf(f, "\n");

		/* run interfaces */
		p = m->ifs;
		fprintf(f, "echo \"Start interfaces...\"\n");
		while (p) {
			s = dinouml_if_build(p);
			fprintf(f, "%s\n", s);
			p = p->next;
		}
		fprintf(f, "\n");

		/* run user commands */
		l = m->run;
		fprintf(f, "echo \"User commands...\"\n");
		while (l) {
			fprintf(f, "%s\n", l->val);
			l = l->next;
		}
		fprintf(f, "\n");

		fprintf(f, "echo \"Finish!\"\n");
		fclose(f);

		/* build overlay %s.tar */
		snprintf(buf, sizeof(buf),
			"tar -C%s/overlay_%s -cf \"%s/tmp/%s.tar\" .",
			D->dir, m->name, D->dir, m->name);
		i = 0;
		argv[i++] = "sh";
		argv[i++] = "-c";
		argv[i++] = buf;
		argv[i++] = NULL;
		dinouml_fork(argv, 1, NULL, NULL, NULL);

		/* swap */
		fprintf(stderr, "o Build swap file...\n");
		dinouml_build_swap(m);

		/* check root fs */
		snprintf(root_tmp, sizeof(root_tmp), "%s/%s",
			D->base, m->root);
		ssf = fopen(root_tmp, "r");
		if (!ssf) {
			printf("WARN: Cannot find %s, so I cannot start machine %s!"
				" Hint: read README!\n",
				root_tmp, m->name);
			m = m->machines;
			continue;
		}
		n = stat(root_tmp, &S);
		if (n == 0)
			root_mtime = S.st_mtime;
		else
			root_mtime = time(NULL);
		fclose(ssf);

		/* unlink cow file */
		/* TODO this must have a config option. */
		snprintf(buf, sizeof(buf), "%s/tmp/%s.COW", D->dir, m->name);
		n = stat(buf, &S);
		if (n == 0) {
			cow_mtime = S.st_mtime;
			if ((root_mtime > cow_mtime) || (S.st_size == 0)) {
				printf("WARN: Remove %s file because is older than root or empty!\n",
					buf);
				unlink(buf);
			}
		}

		/* start switches */
		p = m->ifs;
		while (p) {
			dinouml_switch_start(p, D->dir);
			p = p->next;
		}

		/* run! */
		fprintf(stderr, "o Start machine %s...\n", m->name);
		snprintf(tmp, sizeof(tmp), "tmp/%s.TMP",
			m->name);
		/* build sh file to start machine */
		snprintf(ss, sizeof(ss), "%s/tmp/%s.sh",
			D->dir, m->name);
		ssf = fopen(ss, "w");
		if (ssf == NULL) {
			fprintf(stderr, "Cannot create startul file!\n");
			return -1;
		}
		fprintf(ssf, "#!/bin/sh\n\n");
		fprintf(ssf, "set -e\n\n");

		if (m->tmp > 0) {
			fprintf(stderr, "o Mount TMP dir [%s]...\n", tmp);
			dinouml_build_dir(D->dir, tmp);

			fprintf(ssf, "mount none %s/tmp/%s.TMP -t tmpfs -o mode=1777,size=50M\n\n",
				D->dir, m->name);

			fprintf(ssf, "export TMP=\"%s/tmp/%s.TMP\"\n",
				D->dir, m->name);
		}

		if (strcmp(m->nice, "0") != 0)
			fprintf(ssf, "nice -%s", m->nice);

		fprintf(ssf, " %s/%s %s",
			D->base, m->kernel, m->para);

		fprintf(ssf, " mem=%ldM", m->mem);

		fprintf(ssf, " ubdas=%s/tmp/%s.COW,%s/%s",
			D->dir, m->name, D->base, m->root);

		if (m->swap > 0)
			fprintf(ssf, " ubdbs=%s/tmp/%s.SWAP",
				m->dir, m->name);
		fprintf(ssf, " ubdcrs=%s/tmp/overlay.tar",
			D->dir);
		fprintf(ssf, " ubddrs=%s/tmp/%s.tar",
			D->dir, m->name);
		if (m->port != NULL)
			fprintf(ssf, " ssl0=port:%s",
				m->port);
		fprintf(ssf, " umid=%s-%s",
			D->topo, m->name);
		dinouml_host_ifs(m, ssf);

		if (m->tmp > 0)
			fprintf(ssf, "\numount %s/tmp/%s.TMP\n",
				D->dir, m->name);

		fclose(ssf);

		i = 0;
		argv[i++] = "sh";
		argv[i++] = ss;
		argv[i++] = NULL;
		snprintf(log1, sizeof(log1), "%s/tmp/%s.con1", D->dir, m->name);
		snprintf(log2, sizeof(log2), "%s/tmp/%s.con2", D->dir, m->name);
		snprintf(umlpid, sizeof(umlpid), "%s/tmp/%s.pid", D->dir, m->name);
		dinouml_fork(argv, 0, log1, log2, umlpid);

		m = m->machines;
	}

	return 0;
}

static int dinouml_stop(struct DinoUML *D)
{
	struct DinoUML	*m;
	char		*argv[10], topo[256];
	int		i;

	fprintf(stderr, "o Stopping topology %s (%s)...\n",
		D->topo, D->dir);

	m = D->machines;
	while (m) {
		i = 0;
		argv[i++] = "uml_mconsole";
		snprintf(topo, sizeof(topo), "%s-%s", D->topo, m->name);
		argv[i++] = topo;
		argv[i++] = "cad";
		argv[i++] = NULL;

		fprintf(stderr, "o Stopping %s...\n", m->name);
		dinouml_fork(argv, 1, NULL, NULL, NULL);

		m = m->machines;
	}

	/*
	sleep(x);
	kill all machines
	*/

	dinouml_switch_stopall(D);

	return 0;
}

int main(int argc, char *argv[])
{
	char		*dinouml_base = "/data/DinoUML";
	char		*cmd, *topology, *machine, *hosts;
	char		*x;
	struct DinoUML	*D, *M;
	int len;

	dinouml_copy();

	if (argc < 3) {
		fprintf(stderr, "Usage: dinouml topology cmd\n");
		return 1;
	}

	x = getenv("DINOUML_BASE");
	if (x)
		dinouml_base = x;

	topology = argv[1];
	cmd = argv[2];

	/* fix topology para */
	len = strlen(topology);
	while (topology[len - 1] == '/') {
		topology[len - 1] = '\0';
		len--;
	}

	/* check topology */
	D = dinouml_parse(NULL, topology, "main");
	if (!D)
		return 4;

	/* parse hosts */
	hosts = strdup(D->hosts);
	machine = strtok(hosts, " \t");
	while (machine) {
		M = dinouml_parse(D, topology, machine);
		if (M == NULL) {
			fprintf(stderr, "Error loading conf [%s/%s]!\n",
				topology, machine);
			return 5;
		}

		machine = strtok(NULL, " \t");
	}
	free(hosts);

	if (strncmp(cmd, "dump", 4) == 0) {
		dinouml_dump(D);
	} else if (strncmp(cmd, "run", 3) == 0) {
		dinouml_run(D);
	} else if (strncmp(cmd, "start", 3) == 0) {
		dinouml_run(D);
	} else if (strncmp(cmd, "stop", 4) == 0) {
		fprintf(stderr, "Not yet implemented reliably. You must manually clean all stuff.\n");
		dinouml_stop(D);
	} else {
		fprintf(stderr, "Usage: dinouml topology_dir dump|run|stop\n");
	}

	return 0;
}
